from typing import Iterable, List, Tuple, Union

import numpy
import shapely.geometry
from numpy import array
from scipy.spatial import Voronoi
from shapely.geometry import LineString, MultiLineString, MultiPolygon, Polygon
from shapely.ops import unary_union
from warg import Number

from jord.shapely_utilities.polygons import iter_polygons, polygon_has_interior_rings

__all__ = ["find_centerline"]


def find_centerline(
    input_geometry: Union[Polygon, MultiPolygon], step_size: Number = 0.5
) -> MultiLineString:
    """


    :param input_geometry: input geometry
    :type input_geometry: Union[Polygon, MultiPolygon]
    :param step_size: densify the input geometry border by placing additional points at this distance, in meters
    :type step_size: Optional[float]
    :raises NotImplementedError: input geometry is not of type Union[Polygon, MultiPolygon]
    """

    if not (
        isinstance(input_geometry, Polygon) or isinstance(input_geometry, MultiPolygon)
    ):
        raise NotImplementedError

    step_size = abs(step_size)

    ext_xy = input_geometry.envelope.exterior.xy
    _min_x, _min_y = (
        int(min(ext_xy[0])),
        int(min(ext_xy[1])),
    )

    vertices, ridges = get_voronoi_vertices_and_ridges(
        input_geometry, step_size, minx=_min_x, miny=_min_y
    )

    lines = []
    for ridge in ridges:
        if ridge_is_finite(ridge):
            starting_point = create_point_with_restored_coordinates(
                x=vertices[ridge[0]][0],
                y=vertices[ridge[0]][1],
                _min_x=_min_x,
                _min_y=_min_y,
            )
            ending_point = create_point_with_restored_coordinates(
                x=vertices[ridge[1]][0],
                y=vertices[ridge[1]][1],
                _min_x=_min_x,
                _min_y=_min_y,
            )
            linestring = LineString((starting_point, ending_point))

            if linestring_is_within_input_geometry(linestring, input_geometry):
                lines.append(linestring)

    if len(lines) < 2:
        raise Exception("too few ridges")

    return MultiLineString(lines=unary_union(lines))


def get_voronoi_vertices_and_ridges(
    _input_geometry: shapely.geometry.base.BaseGeometry,
    _step_size: float,
    minx: float,
    miny: float,
) -> Tuple[numpy.ndarray, List[List[int]]]:
    borders = densify_border(_input_geometry, _step_size, minx, miny)

    voronoi_diagram = Voronoi(borders)
    return voronoi_diagram.vertices, voronoi_diagram.ridge_vertices


def ridge_is_finite(ridge: Iterable) -> bool:
    return -1 not in ridge


def create_point_with_restored_coordinates(
    x: float, y: float, _min_x: float, _min_y: float
) -> Tuple[float, float]:
    return x + _min_x, y + _min_y


def linestring_is_within_input_geometry(
    linestring: LineString, input_geometry: shapely.geometry.base.BaseGeometry
) -> bool:
    return linestring.within(input_geometry) and len(linestring.coords[0]) > 1


def densify_border(
    _input_geometry: shapely.geometry.base.BaseGeometry,
    _step_size,
    minx: float,
    miny: float,
) -> numpy.ndarray:
    polygons = iter_polygons(_input_geometry)
    points = []
    for polygon in polygons:
        points += interpolated_boundary(polygon.exterior, _step_size, minx, miny)
        if polygon_has_interior_rings(polygon):
            for interior in polygon.interiors:
                points += interpolated_boundary(
                    interior, _step_size, minx=minx, miny=miny
                )

    return array(points)


def interpolated_boundary(
    boundary: shapely.geometry.base.BaseGeometry,
    _step_size: float,
    minx: float,
    miny: float,
) -> List[Tuple[float, float]]:
    line = LineString(boundary)

    return (
        [get_coordinates_of_first_point(line, minx, miny)]
        + get_coordinates_of_interpolated_points(
            line, _step_size, min_x=minx, min_y=miny
        )
        + [get_coordinates_of_last_point(line, minx=minx, miny=miny)]
    )


def create_point_with_reduced_coordinates(
    x: float, y: float, _min_x: float, _min_y: float
) -> Tuple[float, float]:
    return x - _min_x, y - _min_y


def get_coordinates_of_first_point(
    linestring: LineString, minx: float, miny: float
) -> Tuple[float, float]:
    return create_point_with_reduced_coordinates(
        x=linestring.xy[0][0], y=linestring.xy[1][0], _min_x=minx, _min_y=miny
    )


def get_coordinates_of_last_point(
    linestring: LineString, minx: float, miny: float
) -> Tuple[float, float]:
    return create_point_with_reduced_coordinates(
        x=linestring.xy[0][-1], y=linestring.xy[1][-1], _min_x=minx, _min_y=miny
    )


def get_coordinates_of_interpolated_points(
    linestring: LineString, _step_size: Number, min_x: float, min_y: float
) -> List[Tuple[float, float]]:
    interpolation_distance = _step_size
    intermediate_points = []
    while interpolation_distance < linestring.length:
        point = linestring.interpolate(interpolation_distance)
        reduced_point = create_point_with_reduced_coordinates(
            x=point.x, y=point.y, _min_x=min_x, _min_y=min_y
        )
        intermediate_points.append(reduced_point)
        interpolation_distance += _step_size

    return intermediate_points


if __name__ == "__main__":

    def uashdua():
        from shapely.geometry import Polygon
        from jord.qlive_utilities import AutoQliveClient

        polygon = Polygon([[0, 0], [0, 4], [4, 4], [4, 0]])

        with AutoQliveClient() as qlive:
            qlive.add_shapely_geometry(polygon)
            qlive.add_shapely_geometry(find_centerline(polygon))

    uashdua()



from pathlib import Path
from typing import Any, Generator

import pandas
import shapely
from shapely import wkb

__all__ = ["load_wkbs_from_csv", "csv_wkt_generator"]


def load_wkbs_from_csv(
    csv_file_path: Path, geometry_column: str = "Shape"
) -> pandas.DataFrame:
    """
    Well-Known Text
    """

    return pandas.read_csv(str(csv_file_path))[geometry_column].apply(wkb.loads)


def csv_wkt_generator(
    csv_file_path: Path, geometry_column: str = "Shape"
) -> Generator[shapely.geometry.base.BaseGeometry, Any, None]:
    """

    :param csv_file_path:
    :param geometry_column:
    :return:
    """
    import pandas

    for idx, g in pandas.read_csv(
        str(csv_file_path), usecols=[geometry_column]
    ).iterrows():
        yield wkb.loads(g)  # g is pandas Series?



from typing import Any, Mapping

import osmnx
import shapely
from draugr.numpy_utilities import positive_int_hash
from networkx import MultiDiGraph, MultiGraph

__all__ = [
    "assertive_add_edge",
    "IllegalLoopException",
    "IllegalDuplicateEdgeException",
]

from warg import ensure_existence, recursive_flatten

from jord import PROJECT_APP_PATH


class IllegalLoopException(Exception): ...


class IllegalDuplicateEdgeException(Exception): ...


def assertive_add_edge(
    graph: MultiGraph,
    u: int,
    v: int,
    key: int,
    attributes: Mapping[str, Any],
    *,
    allow_loops: bool = True,
    allow_duplicates: bool = False,
) -> None:
    """

    :param graph: The Graph
    :type graph: MultiDiGraph
    :param u: from node id
    :type u: int
    :param v: to node id
    :type v: int
    :param key: id of edge
    :type key: int
    :param attributes: attributes of edge
    :type attributes: Mapping[str, Any]
    :param allow_loops: Allow loops
    :type allow_loops: bool
    :param allow_duplicates: Allow duplicate edges
    :type allow_duplicates: bool
    :return: None
    """
    if not allow_loops:
        if u == v:
            raise IllegalLoopException(f"{u} == {v}")

    assert isinstance(u, int), f"{u=} is not int, but {type(u)=}"
    assert isinstance(v, int), f"{v=} is not int, but {type(v)=}"

    assert graph.has_node(u)
    assert graph.has_node(v)

    if not allow_duplicates and graph.has_edge(u, v, key):
        if graph.has_edge(u, v, key):
            raise IllegalDuplicateEdgeException(
                f"Graph already contains the edge ({u} -> {v}) with {key=}"
            )

    graph.add_edge(u, v, key=key, **attributes)


def assertive_add_shapely_node(
    graph: MultiDiGraph, u: int, point: shapely.Point, **kwargs
) -> None:
    """
    Add a shapely point based node to the graph.


    :param graph: The Graph
    :type graph: MultiDiGraph
    :param u: Node id
    :type u: int
    :param point:
    :type point: shapely.Point
    :param kwargs: Attributes of node
    :return: None
    """
    assert isinstance(u, int)

    graph.add_node(
        u,
        x=float(point.x),
        y=float(point.y),
        id=u,
        **kwargs,
    )


def network_to_osm_xml(network: MultiDiGraph) -> bytes:
    osm_cache_path = ensure_existence(PROJECT_APP_PATH.site_cache) / "to_upload.osm"

    edge_tags = set(recursive_flatten([edge.keys() for edge in network.edges.values()]))
    node_tags = set(recursive_flatten([node.keys() for node in network.nodes.values()]))

    osmnx.settings.useful_tags_way = edge_tags
    osmnx.settings.useful_tags_node = node_tags

    osmnx.save_graph_xml(G=network, filepath=osm_cache_path)

    with open(osm_cache_path, "rb") as f:
        return f.read()


def compute_node_id(point: shapely.geometry.Point) -> int:
    return positive_int_hash(
        ",".join(str(c) for c in (point.x, point.y, point.z if point.has_z else None))
    )



from typing import Any, Generator

# noinspection PyUnresolvedReferences
from qgis.core import QgsLayerTreeGroup, QgsLayerTreeLayer

__all__ = ["recurse_layers"]


def recurse_layers(group: QgsLayerTreeGroup) -> Generator[Any, None, None]:
    # QgsLayerTreeLayer
    for layer in group.children():
        if isinstance(layer, QgsLayerTreeGroup):
            yield from recurse_layers(layer)
        elif isinstance(layer, QgsLayerTreeLayer):
            yield layer



import datetime
import logging
from typing import (
    Any,
    Collection,
    Generator,
    Iterable,
    List,
    Mapping,
    Optional,
    Tuple,
    Union,
)

import numpy
import pandas
import shapely
from pandas.core.generic import NDFrame

ADD_STRING_LEN = True
NUM_MB16_CHARS = 16777216

__all__ = [
    "solve_qgis_type",
    "solve_type_configuration",
    "solve_attribute_uri",
    "solve_field_uri",
    "to_string_if_not_of_exact_type",
]

logger = logging.getLogger(__name__)


def solve_qgis_type(
    d: Any, floating_point_length: int = 20, floating_point_precision: int = 8
) -> Optional[str]:
    """
    Does not support size/length yet...

    QGIS Available Field types:
    ________________           - Provider type ( MemoryLayer) - implemented

    Whole Number (integer) - integer - X
    Decimal Number (real) - double - X
    Text (string) - string - X
    Date - date - X
    Time - time - X
    Date & Time - datetime - X
    Whole Number ( ... llint - 16bit) - int2 - O
    Whole Number (integer - 32bit) - int4 - O
    Whole Number (integer - 64bit) - int8 - O
    Decimal Number (numeric) - numeric - O
    Decimal Number (decimal) - decimal - O
    Decimal Number (real) - real - O
    Decimal Number (double) - double precision - O
    Text, unlimited length (text) - text - X
    Boolean - boolean  - X
    Binary Object (BLOB) - binary - X
    String List - stringlist - X
    Integer List - integerlist - O
    Decimal (double) List - doublelist - O
    Integer (64 bit) List - integer64list - O
    Map - map - O
    Geometry - geometry - O

    :param floating_point_precision:
    :param floating_point_length:
    :param d:
    :return:
    """
    if d is None:
        return None

    if not isinstance(d, bool):
        if isinstance(d, int):
            return "integer"

        elif isinstance(d, float):  # and (math.isnan(d)):
            return f"double"  # ({floating_point_length,floating_point_precision})" #'real' #"double"

        elif isinstance(d, bytes):
            return "binary"

        elif isinstance(d, (list, tuple, set)):  # ASSUME IS STRINGS
            return "stringlist"

        elif isinstance(
            d,
            (
                datetime.datetime
                #    , pandas.NaT
            ),
        ):
            return "datetime"

        elif isinstance(d, datetime.date):
            return "date"

        elif isinstance(d, datetime.time):
            return "time"

        elif isinstance(d, shapely.Polygon):
            return "geometry"  # "polygon"

        elif isinstance(d, shapely.MultiPolygon):
            return "geometry"  # "multipolygon"

        elif isinstance(d, shapely.LineString):
            return "geometry"  # "linestring"

        elif isinstance(d, shapely.MultiLineString):
            return "geometry"  # "multilinestring"

        elif isinstance(d, shapely.MultiPoint):
            return "geometry"  # "multipoint"

        elif isinstance(d, shapely.Point):
            return "geometry"  # "point"

        elif isinstance(d, str):
            if False:
                if (
                    ADD_STRING_LEN
                ):  # WARNING! Shapefiles have a limitation of maximum 254 characters per field
                    return f"string({min(max(len(d) * 16, 255), NUM_MB16_CHARS)})"  # 16x buffer for large strings
            else:
                return "text"
        else:
            if False:
                logger.error(f"Could not solve type {type(d)=}, {d.__class__.__name__}")

    if numpy.isnan(d):
        return None

    if isinstance(d, NDFrame) and d.size == 0:
        return None

    if pandas.isna(d):
        return None

    if isinstance(d, bool):
        if False:
            if ADD_STRING_LEN:
                return "string(255)"  # True, False (5)
        else:
            if False:
                logger.error(f"Fallback solve type {type(d)=}, {d.__class__.__name__}")
            return "boolean"

    if False:
        return "text"

    return "string"


def solve_attribute_uri(
    attr_type_sampler: Generator,
    columns: Collection,
    floating_point_length: int = 20,
    floating_point_precision: int = 8,
) -> Tuple[Mapping[str, str], Mapping[str, str], int]:

    sample_row = next(attr_type_sampler)
    num_cols = len(sample_row)

    fields = {}
    field_type_configuration = {}
    for data_key, data_value in sample_row.items():
        data_type = solve_qgis_type(data_value)

        if fields.get(data_key) is None:
            fields[data_key] = data_type

        if data_type:
            if field_type_configuration.get(data_key) is None:
                field_type_configuration[data_key] = solve_type_configuration(
                    data_value, data_key, columns
                )

    for row in attr_type_sampler:
        _solved = True
        for data_key in fields.keys():
            if fields.get(data_key) is None:
                _solved = False

        if _solved:
            break

        for data_key, data_value in row.items():
            data_type = solve_qgis_type(data_value)

            if fields.get(data_key) is None:
                fields[data_key] = data_type

            if data_type:
                if field_type_configuration.get(data_key) is None:
                    field_type_configuration[data_key] = solve_type_configuration(
                        data_value, data_key, columns
                    )

    for f in list(fields.keys()):
        if f not in field_type_configuration:
            field_type_configuration[f] = None

    return field_type_configuration, fields, num_cols


def solve_field_uri(
    field_type_configuration: Mapping, fields: Mapping, uri: str
) -> str:
    uri = str(uri).rstrip("&")

    for k, v in fields.items():
        uri += f"&field={k}:{v}"
        if field_type_configuration is not None and k in field_type_configuration:
            c = field_type_configuration[k]
            if c:
                uri += f"({c})"

    return uri


def to_string_if_not_of_exact_type(
    gen: Iterable, type_: Iterable[type] = (int, float, str, bool, datetime.datetime)
) -> Union[str, Any]:
    """

    :param type_: Type for testing against
    :param gen: The iterable to be converted
    :return:
    """
    if not isinstance(type_, Iterable):
        type_ = [type_]

    for v in gen:
        if v is None:
            yield None
        elif isinstance(v, Collection):
            if False:
                yield list(
                    str(v_) if all([type(v_) != t for t in type_]) else v_ for v_ in v
                )  # TODO: SHOULD ALSO BE CONVERTED?
            else:
                yield v
        elif pandas.isna(v):
            yield None
        elif all([type(v) != t for t in type_]):  # UNKNOWN TYPE
            yield str(v)
        else:
            yield v


def solve_type_configuration(
    data_value: Any,
    data_key: Optional[str],
    columns: Optional[List],
    allocation_multiplier: Optional[int] = 2,
    floating_point_length: int = 20,
    floating_point_precision: int = 8,
) -> Optional[str]:
    """

      :param floating_point_length:
    :param floating_point_precision:
    :param data_value:
    :param data_key:
    :param columns:
    :param allocation_multiplier:
    :return:
    """
    if isinstance(data_value, float):
        return f"{floating_point_length},{floating_point_precision}"

    if isinstance(data_value, str):
        a = len(data_value)

        if data_key and columns:
            max_len = a

            if isinstance(columns, Iterable):
                for cols in columns:
                    c = cols[data_key]

                    if isinstance(c, str):
                        max_len = max(max_len, len(c))

            a = max_len

        if allocation_multiplier:
            a *= allocation_multiplier

        a = max(a, 255)
        return str(a)

    return None



# noinspection PyUnresolvedReferences
from qgis.core import QgsEditorWidgetSetup


HIDDEN_WIDGET = QgsEditorWidgetSetup("Hidden", {})

CHECKBOX_WIDGET = QgsEditorWidgetSetup(
    "CheckBox",
    {"CheckedState": "True", "UncheckedState": "False", "TextDisplayMethod": 0},
)

NULLABLE_CHECKBOX_WIDGET = QgsEditorWidgetSetup(
    "CheckBox",
    {
        "CheckedState": "True",
        "UncheckedState": "False",
        "AllowNullState": True,
        "TextDisplayMethod": 0,
    },
)

UNIQUE_VALUES_WIDGET = QgsEditorWidgetSetup(
    "UniqueValues",
    {"Editable": True},
)

COLOR_WIDGET = QgsEditorWidgetSetup(
    "Color",
    {},
)


def make_range_widget(
    min_value: float = 0.0, max_value: float = 360.0, step_size: float = 0.5
) -> QgsEditorWidgetSetup:
    """

    :param min_value:
    :param max_value:
    :param step_size:
    :return:
    """
    return QgsEditorWidgetSetup(
        "Range",
        {
            "Min": min_value,
            "Max": max_value,
            "Step": step_size,
            "UseCompleter": False,
            "NofColumns": 1,
            "CompleterMatchFlags": 2,
        },
    )



from typing import Any, Collection, List, Mapping, Optional

from pandas import DataFrame

__all__ = ["df_to_columns", "columns_to_df"]


def df_to_columns(
    shape_df: DataFrame, ignored_columns: Optional[Collection] = None
) -> List[Mapping[str, Any]]:
    columns = []

    for key, c in shape_df.iterrows():
        o = {"key": key, **dict(c)}

        if ignored_columns:
            for p in ignored_columns:
                o.pop(p)

        columns.append(o)

    return columns


def columns_to_df(columns: List[Mapping[str, Any]]) -> DataFrame: ...
